(** * Definition of the floor function *)
(** Contributed by David Baelde, 2011 *)

Require Import Rplus.
Require Import RpRing.
Require Import Uprop.
Require Import Setoid.

Open Scope Rp_scope.

Definition floor : Rp -> nat := fun r =>
  if isle_dec U1 (frac r) then S (int r) else int r.

Lemma floor_int : forall x, frac x < 1%U -> floor x = int x.
intros; unfold floor; rewrite if_else; auto.
Save.
Hint Resolve floor_int.

Lemma floor_int_equiv : forall x, frac x < 1%U <-> floor x = int x.
split; auto.
unfold floor; destruct (isle_dec U1 (frac x)); intros; auto with arith.
exfalso; omega.
Save.


Lemma floor_S_int : forall x, 1%U <= frac x -> floor x = S (int x).
intros; unfold floor; rewrite if_then; auto.
Save.
Hint Resolve floor_S_int.

Lemma floor_S_int_equiv : forall x, 1%U <= frac x <-> floor x = S (int x).
split; auto.
unfold floor; destruct (isle_dec U1 (frac x)); intros; auto with arith.
exfalso; omega.
Save.


Lemma floor_le_S_int : forall x, (floor x <= S (int x))%nat.
Proof.
  intro x; unfold floor.
  destruct (isle_dec U1 (frac x)); simpl; auto.
Qed.

Lemma int_le_floor : forall x, (int x <= floor x)%nat.
Proof.
  intro x; unfold floor.
  destruct (isle_dec U1 (frac x)); simpl; auto.
Qed.

Hint Resolve floor_le_S_int int_le_floor.

Lemma floor_le : forall x, N2Rp (floor x) <= x.
Proof.
  intro; unfold floor.
  destruct (isle_dec 1 (frac x)); [|auto].
  apply Rple_intro_norm.
  unfold norm; split.
  reflexivity.
  split; auto.
Qed.

Hint Resolve floor_le.

Lemma floor_N2Rp : forall n, floor (N2Rp n) = n.
Proof.
  intros; unfold floor.
  rewrite fracN2Rp.
  rewrite if_else; auto.
Qed.

Hint Resolve floor_N2Rp.

Lemma floor_le_plus : forall x y, (floor x + floor y <= floor (x+y))%nat.
Proof.
  intros.
  destruct (isle_dec (frac x) ([1-]frac y)).
  (* 1. frac x <= 1-frac y *)
  rewrite Rpplus_simpl_le; trivial.
  unfold floor; destruct (isle_dec U1 (frac x)); 
  repeat rewrite frac_simpl; repeat rewrite int_simpl.
  (* 1.1 1 <= frac x *)
  rewrite if_else.
  rewrite if_then; auto.
  rewrite o0; auto.
  apply Olt_notle; auto.
  apply Ole_lt_trans with 0; auto.
  apply Uinv_le_simpl; rewrite <- o; auto.
  (* 1.2 frac x < 1 *)
  destruct (isle_dec U1 (frac y)).
  rewrite if_then; try omega.
  rewrite <- o0; auto.
  destruct (isle_dec U1 (frac x + frac y)); auto with arith.
  (* 2. int part is incremented in plus *)
  rewrite Rpplus_simpl_lt; auto.
  unfold floor at 3; simpl frac; simpl int.
  destruct (isle_dec U1 (frac x & frac y)).
    (* 2.1 *)
    apply le_trans with ((S (int x)) + (S (int y)))%nat; auto with zarith.
    apply plus_le_compat; auto.
    destruct (isle_dec U1 (frac x)).
    rewrite floor_S_int; auto.
    rewrite floor_int; auto.
    apply Ole_lt_trans with (frac x & frac y)%U.
    transitivity (1 & frac y)%U; auto.
    apply notUle_lt; auto.
    rewrite floor_int; auto.
    apply le_trans with (int x + S (int y))%nat; auto with arith; try omega.
Qed.

Add Morphism floor with signature Ole ==> le as floor_le_compat.
intros.
rewrite Rple_simpl in H.
destruct H as [H|[(H1,H2)|H]].
apply le_trans with (S (int x)); trivial.
apply le_trans with (int y); trivial.
unfold floor at 1; destruct (isle_dec U1 (frac x)).
rewrite floor_S_int; auto with zarith.
transitivity (frac x); auto.
apply le_trans with (int y); auto with zarith.
destruct H as (H1,(H2,H3)).
rewrite floor_int.
rewrite floor_S_int; auto with zarith.
rewrite H3; auto.
Qed.
Hint Immediate floor_le_compat.

Add Morphism floor with signature Oeq ==> eq as floor_eq_compat.
intros; apply le_antisym; apply floor_le_compat; auto.
Save.

Lemma floor_gt_S : forall x, x < S (floor x).
Proof.
  intros.
  destruct (isle_dec 1 (frac x)) as [H|H].
  rewrite floor_S_int; auto.
  apply Ole_lt_trans with (N2Rp (S (int x))); auto.
  rewrite floor_int; auto.
  rewrite Rpplus_int_frac at 1.
  rewrite Rpplus_sym; apply Olt_le_trans with (R1 + (int x)); auto.
  apply Rpplus_lt_compat_left; auto.
  rewrite <- U2Rp1_R1; auto.
Qed.
Hint Resolve floor_gt_S.

Lemma floor_gt : forall x, x < 1 + (floor x).
Proof.
  intros.
  apply Olt_eq_compat with x (N2Rp (S (floor x))); auto.
Qed.
Hint Resolve floor_gt.

(** A weaker version useful for auto *)
Lemma floor_ge : forall x, x <= 1 + (floor x).
auto.
Save.
Hint Resolve floor_ge.

Lemma floor_lt_simpl : forall x y, x + 1 <= y -> x < floor y.
Proof.
  intros.
  apply Olt_eq_compat with (x + 1 - 1) (1 + floor y - 1); auto.
  rewrite H; auto.
  apply Rpminus_lt_compat_left; auto.
  transitivity (x + 1); auto.
Qed.
Hint Resolve floor_lt_simpl.


Lemma floor_le_simpl : forall x y, x + 1 <= y -> x <= floor y.
Proof.
  intros.
  apply (floor_lt_simpl x y H).
Qed.
Hint Resolve floor_le_simpl.

